# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Base classes for prior classifiers."""

import math
from abc import ABC
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Any, Callable, Iterable, List, Sequence, Union

from .actions import Action
from .color import Color


class ScalingMode(Enum):
    """No scaling at all, preserve the raw score"""

    IDENTITY = auto()

    """Any values less than 0 are set to 0, any values larger than 1 are set to 1"""
    CLAMP = auto()

    """Map values to the [0,1] range"""
    LINEAR = auto()

    """Map the logarithm of the values to the [0,1] range"""
    LOG = auto()

    def scale(self, values: Sequence[float]) -> List[float]:
        """Scales a list of scores according to a given mode"""
        if self == ScalingMode.IDENTITY:
            return list(values)

        if self == ScalingMode.CLAMP:
            return [min(1, max(0, score)) for score in values]

        minimum, maximum = min(values), max(values)

        if minimum == maximum or len(values) <= 1:
            return [1.0] * len(values)

        if self == ScalingMode.LINEAR:
            return [(score - minimum) / (maximum - minimum) for score in values]

        return [(math.log(score) - math.log(minimum)) / (math.log(maximum) - math.log(minimum)) for score in values]


@dataclass
class Classifier(ABC):
    """
    Base class for all prior classifiers. Do not use,
    refer instead to :class:`ElementClassifier` or :class:`ViewClassifier`.
    """

    name: str
    enabled: bool = True
    callback: Any = None


@dataclass
class ElementClassifier(Classifier):
    """
    Classifies a set of elements. The callback will receive a list of
    elements that have have tags for all tags given in subset.

    The callback either returns a sublist of elements, or a list of tuples
    mapping element to a numeric score.

    If the callback is doing multi-class prediction, then the output should be
    a dictionary mapping class name to a sublist or list of tuples described above.
    The prediction results would be stored as <classifier_name>__<class name>.

    If highlight is True, highlight every element returned by this classifier.
    If highlight is a float x, highlight every element with a score larger than x.
    If highlight is an int N, highlight the top N scoring elemnets.
    """

    action: Action = None
    highlight: Union[float, bool] = False
    mode: ScalingMode = ScalingMode.CLAMP
    highlight_color: Color = Color.from_str("#5A1911")
    subset: Union[str, Iterable[str]] = "all"
    result_type: type = float


@dataclass
class ViewClassifier(Classifier):
    """
    Classifies a given view. The callback will receive a view
    and return an iterable of string tags.
    """


def _active_element_filter_func(elements, workflow):
    actives = set(workflow.js.execute_file(Path("find_active_elements.js")))
    return [elem for elem in elements if elem.wtl_uid in actives]


@dataclass
class ActiveElementFilter(ElementClassifier):
    """
    Returns all elements that are considered active, i.e. interactable in some way.
    Will also add a boolean `is_active` field to every element's metadata.
    """

    name: str = "is_active"
    callback: Callable = _active_element_filter_func
    result_type: type = bool



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Abstraction layer for a screenshot of a site"""

from __future__ import annotations

import io
import os
import time
from math import ceil
from pathlib import Path
from typing import Dict, List

from PIL import Image
from selenium.webdriver.remote.webdriver import WebDriver

from .color import Color
from .geometry import Point, Rectangle
from .graphics import draw_rect, draw_text
from .javascript import JavascriptWrapper


class Screenshot:
    """Abstraction layer for a screenshot of a site, allowing for various annotations."""

    def __init__(self, name: str, image: Image.Image):
        self.name = name
        self.image = image

    @classmethod
    def capture_viewport(cls, name: str, driver: WebDriver, scale: float = 1.0) -> Screenshot:
        """
        Creates a screenshot of the current viewport of a given webdriver.
        Scales the image by some pixel ratio, if given.
        Uses PIL as a backend.
        """
        page_screenshot_png_bytes = driver.get_screenshot_as_png()
        image = Image.open(io.BytesIO(page_screenshot_png_bytes)).convert("RGB")
        if scale != 1.0:
            new_size = [int(x * scale) for x in image.size]
            image = image.resize(new_size)
        return Screenshot(name, image)

    @classmethod
    def capture(cls, name: str, driver: WebDriver, scale: float = 1.0, max_page_height: int = 0) -> Screenshot:
        """
        Creates a snapshot on the given webdriver under certain conditions.
        """
        js = JavascriptWrapper(driver)
        viewport = js.find_viewport()
        if max_page_height < viewport.height:
            return cls.capture_viewport(name, driver, scale)

        # Capture at least one viewport's height, at most the page's height
        minimum_length = max(int(viewport.height), js.get_full_height())
        page_height = min(max_page_height, minimum_length)
        num_shots = ceil(page_height / viewport.height)
        viewport_screenshots: List[Image.Image] = []
        hidden_elements: Dict[int, str] = js.hide_position_fixed_elements()

        try:
            # We start from bottom, and go up!
            for shot_idx in range(num_shots, -1, -1):
                # Don't hide elements at the top
                if shot_idx == 0 or viewport.height > page_height:
                    js.show_position_fixed_elements(hidden_elements)

                # Scroll, hide sticky elements, and take viewport screenshot
                hidden = Screenshot._scroll_to(js, shot_idx * int(viewport.height))
                hidden_elements.update(hidden)
                viewport_screenshot = cls.capture_viewport("temp", driver, scale).image

                if shot_idx * viewport.height > page_height - viewport.height:
                    new_viewport_area = (
                        0,
                        0,
                        viewport.width,
                        page_height - shot_idx * viewport.height,
                    )
                    viewport_screenshot = viewport_screenshot.crop(new_viewport_area)

                # Append screenshot of the current viewport at the top of the image
                viewport_screenshots.insert(0, viewport_screenshot)
        finally:
            # Unhide everything and scroll back to the top
            js.show_position_fixed_elements(hidden_elements)

        total_height = sum(image.height for image in viewport_screenshots)
        final_screenshot = Image.new("RGB", (viewport_screenshots[0].width, total_height))

        offset = 0
        for viewport_screenshot in viewport_screenshots:
            final_screenshot.paste(viewport_screenshot, (0, offset))
            offset += viewport_screenshot.height

        return Screenshot(name, final_screenshot)

    @classmethod
    def load(cls, name: str, path: Path) -> Screenshot:
        return cls(name, Image.open(str(path)))

    def save(self, path: Path, suffix: str = ""):
        """Saves screenshot to given path. Filename consists of the screenshot name
        and an optional suffix."""
        os.makedirs(path, exist_ok=True)
        filename = f"{self.name}_{suffix}.png" if suffix else f"{self.name}.png"
        self.image.save(str(path / filename))

    def copy(self, new_name: str) -> Screenshot:
        return Screenshot(new_name, self.image.copy())

    def highlight(self, rect: Rectangle, color: Color, text: str = "", width: int = 1):
        """
        Draws a colored rectangle on the screenshot.
        Can also annotate with a text below the rectangle, if given.
        """
        draw_rect(self.image, rect, color, width)
        if text:
            offset = 1
            self.annotate(Point(rect.x + offset, rect.y + rect.height + offset), color, 12, text)

    def annotate(self, top_left: Point, color: Color, size: int, text: str):
        """
        Writes text with a given color on the screenshot.
        """
        draw_text(self.image, top_left, color, size, text)

    @property
    def size(self) -> Point:
        """Returns a (width, height) Point of the screenshot size in pixels"""
        return Point(self.width, self.height)

    @property
    def height(self) -> int:
        return self.image.height

    @property
    def width(self) -> int:
        return self.image.width

    @staticmethod
    def _scroll_to(js: JavascriptWrapper, y_pos: int, elements: List[str] = None):
        waiting_time = 0.1
        hidden: Dict[int, str] = {}

        js.scroll_to(0, y_pos)
        time.sleep(waiting_time)
        hidden = js.hide_position_fixed_elements(elements=elements if elements else [])
        time.sleep(waiting_time)

        return hidden



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Module for scraping a website and saving contents to file.
"""

import logging
import os.path
from datetime import datetime
from pathlib import Path
from time import sleep
from typing import Callable, Dict, List

import bs4
from selenium.common.exceptions import TimeoutException, UnexpectedAlertPresentException, WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait
from urllib3.util import parse_url

from .config import Config
from .javascript import JavascriptWrapper
from .processtools import TimeoutContext
from .screenshot import Screenshot
from .snapshot import PageSnapshot
from .version import __version__

logger = logging.getLogger("wtl")


class Scraper:
    """
    Used to create web page snapshots using a WebDriver instance.
    """

    def __init__(
        self, driver: WebDriver, config: Config, hide_sticky: bool = True, postload_callbacks: List[Callable] = None
    ):
        self.driver = driver
        self.config = config
        self.hide_sticky = hide_sticky
        self.js = JavascriptWrapper(self.driver, config)
        self.postload_callbacks = postload_callbacks
        self.device_pixel_ratio = self.js.execute_script("return window.devicePixelRatio;") or 1.0

    def scrape_current_page(self) -> PageSnapshot:
        """
        Scrape the page currently open in the driver,
        """
        attempts = self.config.scraping.attempts
        if attempts < 1:
            logger.warning("config.scraping.attempts is set to 0, no snapshot will be made!")

        snapshot = None
        while attempts > 0:
            try:
                attempts -= 1
                snapshot = self._create_snapshot()
                break
            except WebDriverException as e:
                logger.error(e)
                if attempts <= 0:
                    raise

            logger.warning("Scraping failed, reloading the page and trying again...")

            while attempts > 0:
                sleep(1.0)

                try:
                    self.refresh()
                    break
                except WebDriverException as e:
                    attempts -= 1
                    logger.error(e)
                    if attempts <= 0:
                        raise

        return snapshot

    def refresh(self):
        """
        Triggers a refresh of the page.
        Blocks until page is loaded.
        """
        self.driver.refresh()
        if self.config.scraping.disable_animations:
            self.js.disable_animations()
        self.wait_until_loaded()

    def navigate(self, url: str):
        """
        Navigate the scraper's internal webdriver to the URL.
        Blocks until page is loaded.
        """
        try:
            if parse_url(url).scheme is None:
                url = "http://" + url
            self.driver.get(url)
        except TimeoutException:
            raise TimeoutError(f"WebDriver page load timed out on url '{url}'")
        except WebDriverException:
            logger.error(f"URL not valid: {url}")

        if self.config.scraping.disable_animations:
            self.js.disable_animations()
        self.wait_until_loaded()

    def wait_until_loaded(self, timeout: int = None):
        """
        Waits on the webdriver instance to finish loading a page before returning.

        .. note::
            Because of the unending imagination of javascript devs, there may be cases where
            this function returns before the timeout although the page hasn't loaded.
        """
        if timeout is None:
            timeout = self.config.scraping.page_load_timeout

        try:
            # Wait until page is loaded from JS and jQuery perspective
            WebDriverWait(self.driver, timeout).until(self.js.is_page_loaded)
        except TimeoutException:
            logger.warning("Timed out waiting for the page to fully load, proceeding anyway")
        except UnexpectedAlertPresentException:
            logger.warning("Javascript alert noted, trying to proceed")

        sleep(self.config.scraping.wait_loading)

        # Some pages change after scrolling, so simulate some movement
        if self.config.scraping.prescroll:
            self.js.scroll_to(0, 100)
            sleep(self.config.scraping.wait_scroll)
            self.js.scroll_to(0, 9999)
            sleep(self.config.scraping.wait_scroll)
            self.js.scroll_to(0, 0)
            sleep(self.config.scraping.wait_scroll)

    def capture_screenshot(self, name: str, max_page_height: int = 0) -> Screenshot:
        """
        Captures the screenshot of the current rendering in the browser window.
        """
        return Screenshot.capture(
            name=name, driver=self.driver, scale=1 / self.device_pixel_ratio, max_page_height=max_page_height
        )

    def get_page_as_mhtml(self) -> bytes:
        """
        Gets an MHTML representation of the current page, returns it as a bytestring.
        MHTML must be enabled in the browser configuration, otherwise returns None.
        """
        if not self.config.browser.enable_mhtml:
            return None

        folder = Path(os.path.expanduser(self.config.scraping.temp_path))
        os.makedirs(folder, exist_ok=True)
        filename = "temp_mhtml_file.txt"

        try:
            os.remove(folder / filename)
        except OSError:
            pass

        self.js.save_mhtml(filename)

        try:
            with TimeoutContext(n_seconds=self.config.scraping.mhtml_timeout):
                while not os.path.exists(folder / filename):
                    sleep(0.25)
        except TimeoutError:
            logger.error("Failed to get MHTML snapshot within desired time limit!")
            return None

        with open(folder / filename, "rb") as f:
            return f.read()

    def _create_snapshot(self) -> PageSnapshot:
        before = datetime.now()

        logger.debug(f"Page title: {self.driver.title}")

        # Create screenshots if required
        screenshots: Dict[str, Screenshot] = {}
        if self.config.debug.screenshots:
            max_page_height = self.config.scrolling.max_page_height
            screenshots["first"] = self.capture_screenshot("first")
            if max_page_height == 0:
                screenshots["full"] = screenshots["first"].copy("full")
            else:
                screenshots["full"] = self.capture_screenshot("full", max_page_height=max_page_height)

        # Gather element metadata
        elements_metadata = self.js.get_element_metadata() or []
        num_elements = len(elements_metadata)

        # Gather page metadata
        inner_html = self.driver.find_element(By.XPATH, "/html").get_attribute("innerHTML")
        page_source = bs4.BeautifulSoup(f"<!DOCTYPE html><html>{inner_html}</html>", self.config.bs_html_parser)
        page_metadata = {
            "timestamp": before.isoformat(),
            "url": self.driver.current_url,
            "title": self.driver.title,
            "driver": self.driver.name,
            "full_page_size": (self.config.browser.width, self.js.get_full_height()),
            "device_pixel_ratio": self.device_pixel_ratio,
            "num_elements": num_elements,
            "screenshots": list(screenshots.keys()),
            "wtl_version": __version__,
        }

        milliseconds_passed = (datetime.now() - before).total_seconds() * 1000
        if num_elements < 2:
            logger.info("No elements detected on the page!")
        else:
            logger.info(f"Found {num_elements} " f"elements in {milliseconds_passed:.2f}ms")

        mhtml_source = self.get_page_as_mhtml() if self.config.scraping.save_mhtml else None

        # Assemble snapshot
        snapshot = PageSnapshot(
            page_source=page_source,
            page_metadata=page_metadata,
            elements_metadata=elements_metadata,
            screenshots=screenshots,
            mhtml_source=mhtml_source,
        )

        return snapshot



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Contains built-in simple, common goals."""


class N_STEPS:
    """Will return False (continue) exactly n times, and then return True."""

    def __init__(self, n: int):
        assert n >= 0
        self.n = n + 1

    def __call__(self, *_, **__):
        self.n -= 1
        return self.n <= 0


class ONCE(N_STEPS):
    """Will return False (continue) exactly once."""

    def __init__(self):
        super().__init__(n=1)


def FOREVER(*_, **__):
    """Will always return False (continue)."""
    return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Contains common library-level errors"""

from __future__ import annotations

import logging
from functools import wraps

logger = logging.getLogger("wtl")


class Error(Exception):
    """Base class for all WTL-specific exceptions."""

    @classmethod
    def wrapped(cls, func):
        """Wraps a function in a try-except block where AssertionErrors are logged and reraised as Error."""

        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
            except AssertionError as e:
                logger.error(e)
                raise cls(e)
            return result

        return wrapper


class ScrapingError(Error):
    """Any error related to scraping/parsing webpages."""


class ElementNotFoundError(Error):
    """Any error related to finding elements on the page."""


class WindowClosedError(Error):
    """Trying to access a browser instance that was closed by the user"""


class WebDriverSendError(Error):
    """Any error related to custom sending commands to a WebDriver instance"""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
This module contains heuristics for generating selectors.
"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import total_ordering
from typing import Union

import bs4

logger = logging.getLogger("wtl")


@total_ordering
@dataclass(frozen=True)
class Selector:
    """
    Web element selector based on CSS and XPATH.
    You may also specify an identifier (name or ID) of an iframe in which the given element is located.
    The class itself provides no guarantees on whether the selector is unique or even matches anything.
    The ``iframe`` value can be used if this selector refers to an element inside an iframe. Specify as ID or name.
    """

    css: str = "*"
    xpath: str = "/"
    iframe: str = None

    def __lt__(self, other: Selector) -> bool:
        if len(self.css) == len(other.css):
            return self.xpath < other.xpath
        return len(self.css) < len(other.css)

    @classmethod
    def build(cls, bs4_soup: bs4.BeautifulSoup, target: Union[bs4.Tag, int]) -> Selector:
        """
        Compute xpath and css of a ``target`` in a bs4.BeautifulSoup.
        Will be verbose. Use a separate generalizer if you want reusable selectors.
        """
        # Identify target
        if isinstance(target, bs4.Tag):
            element = target
        else:
            options = bs4_soup.find_all(attrs={"wtl-uid": target})
            if len(options) != 1:
                logger.error("Invalid wtl-uid given to Selector.build, returning blank selector")
                return cls(css="bad_wtl_uid_no_matches", xpath="bad_wtl_uid_no_matches")
            element = options[0]

        names, indices = [], []
        child = element if element.name else element.parent

        for parent in child.parents:
            siblings = parent.find_all(child.name, recursive=False)
            index = 1 + siblings.index(child)
            names.append(cls._safe_tag_name(child.name))
            indices.append(-1 if len(siblings) == 1 else index)
            child = parent

        css_components = [
            name if index == -1 else f"{name}:nth-of-type({index})" for (name, index) in zip(names, indices)
        ]
        xpath_components = [name if index == -1 else f"{name}[{index}]" for (name, index) in zip(names, indices)]

        css = ">".join(reversed(css_components)).strip()
        xpath = "/" + "/".join(reversed(xpath_components)).strip()
        return Selector(css=css, xpath=xpath)

    @classmethod
    def _safe_tag_name(cls, name: str) -> str:
        if ":" in name or "=" in name:
            return "*"
        return name



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
An action is an abstraction of some website or browser interaction,
often implemented as a javascript snippet or webdriver API execution applicable for a given view.

Action instances can be incomplete, i.e. they are not initialized with all fields. Call them with
the missing attributes to replace those.
"""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, replace
from time import sleep
from typing import Union

from .color import Color
from .geometry import Point
from .selector import Selector
from .snapshot import Elements, PageElement


@dataclass(frozen=True)
class Action(ABC):
    """
    Base class for all actions. Do not use, refer
    instead to :class:`ElementAction` or :class:`PageAction`.

    Each action must provide an ``execute`` method that performs the required logic.
    """

    def execute(self, workflow):
        pass

    def __call__(self, *args, **kwargs):
        if not args:
            return replace(self, **kwargs)

        # If there is only one non-target element, use the positional argument for that
        assert len(args) == 1
        fields = [x for x in self.__dataclass_fields__ if x != "target"]  # pylint: disable=no-member
        assert len(fields) == 1, "Ambiguous action specialiation, use keyword arguments!"
        return replace(self, **{fields[0]: args[0]})


class Actions(list):
    """Helper class for a list of actions"""

    def by_type(self, tag: type) -> Actions:
        """Returns all actions of the given type."""
        return Actions([action for action in self if isinstance(action, tag)])

    def by_score(self, name: str, limit: float = 0.0) -> Actions:
        """Returns all actions with the given score (metadata entry) greater than the given limit."""
        return Actions(
            [
                action
                for action in self
                if isinstance(action, ElementAction)
                and name in action.target.metadata  # type: ignore
                and action.target.metadata[name] > limit  # type: ignore
            ]
        )

    def by_raw_score(self, name: str, limit: float = 0.0) -> Actions:
        """
        Returns all actions with the given raw score (output from a classifier before scaling)
        greater than the given limit.
        """
        return Actions(
            [
                action
                for action in self
                if isinstance(action, ElementAction)
                and name in action.target.raw_scores  # type: ignore
                and action.target.raw_scores[name] > limit  # type: ignore
            ]
        )

    def by_element(self, element: PageElement) -> Actions:
        """Returns all actions (ElementAction) that act upon the given element."""
        return Actions([action for action in self if isinstance(action, ElementAction) and action.target == element])

    def by_selector(self, selector: Selector) -> Actions:
        """
        Queries the page by the given selector. If at least one element is found,
        return all elements equal to one of those.
        """
        if not self:
            return Actions([])

        element_actions = self.by_type(ElementAction)
        if not element_actions:
            return Actions([])

        tags = set(element_actions[0].target.page.page_source.html.select(selector.css))
        if not tags:
            return Actions([])

        wtl_uids = set(int(x.attrs["wtl-uid"]) for x in tags if "wtl-uid" in x.attrs)
        actions = Actions([action for action in element_actions if action.target.wtl_uid in wtl_uids])

        # Falls back on BS4 tags if selector matches something that hasn't been snapshotted yet
        if not actions:
            actions = Actions([action for action in element_actions if action.target.tag in tags])

        return actions

    def sort_by(self, name: str = None, reverse: bool = False) -> Actions:
        """
        Sorts by a certain action (raw) score. If given name does not exist the element gets (raw) score 0.
        """
        self.sort(key=lambda action: action.target.raw_scores.get(name, 0), reverse=reverse)
        return self


@dataclass(frozen=True)
class ElementAction(Action):
    """
    Base class for all actions that execute on a specific element.
    Can be initialised with a :class:`PageElement` or a css selector string.
    """

    target: Union[PageElement, Selector] = None

    def transformed_to_element(self, elements: Elements) -> ElementAction:
        """Modifies this action with a PageElement corresponding to the stored selector"""
        assert isinstance(self.target, Selector)
        selector = self.target
        tags = elements.by_selector(selector)
        assert tags, f"Failed to perform {self.__class__.__name__}: Selector '{selector.css}' matches no tags"
        page_element = replace(tags[0])
        object.__setattr__(page_element, "selector", selector)
        return replace(self, target=page_element)

    @property
    def selector(self) -> Selector:
        return self.target if isinstance(self.target, Selector) else self.target.selector


@dataclass(frozen=True)
class PageAction(Action):
    """
    Base class for all actions that do not execute on a specific element.
    """


@dataclass(frozen=True)
class Click(ElementAction):
    """
    Simulates a click on the contained element.
    If it isn't clickable, nothing happens.
    """

    def execute(self, workflow):
        with workflow.frame(self.selector.iframe):
            workflow.js.click_element(self.selector)


@dataclass(frozen=True)
class FillText(ElementAction):
    """
    Fills a string by setting the text value in the contained element.
    If it isn't a text field, anything can happen.
    """

    text: str = ""

    def execute(self, workflow):
        with workflow.frame(self.selector.iframe):
            workflow.js.fill_text(self.selector, self.text)


@dataclass(frozen=True)
class Select(ElementAction):
    """
    Selects the given value on a <select> dropdown element.
    If it isn't a select element, anything can happen.
    """

    value: str = ""

    def execute(self, workflow):
        with workflow.frame(self.selector.iframe):
            workflow.js.select(self.selector, self.value)


@dataclass(frozen=True)
class ScrollTo(ElementAction):
    """
    Scrolls the current page to center the given element vertically.
    """

    def execute(self, workflow):
        workflow.smart_scroll_to(self.target.bounds)


@dataclass(frozen=True)
class Highlight(ElementAction):
    """
    Highlights an element by calling workflow.js.highlight(...)
    viewport refers to drawing on a floating viewport-sized canvas. If None, uses default value from config.
    """

    color: Color = Color.from_str("#FFB3C7")
    fill: bool = False
    viewport: bool = None

    def execute(self, workflow):
        viewport = workflow.config.debug.default_canvas_viewport if self.viewport is None else self.viewport
        workflow.js.highlight(selector=self.selector, color=self.color, fill=self.fill, viewport=viewport)


@dataclass(frozen=True)
class Remove(ElementAction):
    """Removes the given element from the DOM."""

    def execute(self, workflow):
        workflow.js.delete_element(self.selector)


@dataclass(frozen=True)
class Annotate(PageAction):
    """
    Writes text on a given page by calling workflow.js.annotate(...).
    viewport refers to drawing on a viewport canvas. If None, uses default value from config.
    """

    location: Point
    color: Color
    size: int
    text: str
    background: Color = Color(0, 0, 0, 0)
    viewport: bool = None

    def execute(self, workflow):
        viewport = workflow.config.debug.default_canvas_viewport if self.viewport is None else self.viewport

        workflow.js.annotate(
            location=self.location,
            color=self.color,
            size=self.size,
            text=self.text,
            background=self.background,
            viewport=viewport,
        )


@dataclass(frozen=True)
class Clear(PageAction):
    """
    Clears all highlights and annotations.
    viewport refers to drawing on a floating viewport-sized canvas. If None, uses default value from config.
    """

    viewport: bool = None

    def execute(self, workflow):
        viewport = workflow.config.debug.default_canvas_viewport if self.viewport is None else self.viewport
        workflow.js.clear_highlights(viewport=viewport)


@dataclass(frozen=True)
class Navigate(PageAction):
    """
    Navigates to a new URL and waits for the page to load.
    If the URL is invalid, you may end up on the browser's error page.
    """

    url: str = ""

    def execute(self, workflow):
        workflow.scraper.navigate(self.url)


@dataclass(frozen=True)
class Revert(PageAction):
    """
    Reverts the state of the Workflow to a previous point in time.
    In effect, resets the underlying web driver and then replays all actions leading up
    to the given view_index. If 0, just perform the initial action of :class:`Navigate`
    to the initial URL.
    """

    view_index: int = 0

    def execute(self, workflow):
        workflow.reset_to(self.view_index)


@dataclass(frozen=True)
class Wait(PageAction):
    """
    Calls :func:`time.sleep` with the given seconds argument.
    """

    seconds: float = 0

    def execute(self, _):
        sleep(self.seconds)


@dataclass(frozen=True)
class WaitForElement(PageAction):
    """
    Checks to see if element at given selector exists on the page.
    Keeps trying indefinitely with a given interval until it succeeds.
    """

    selector: Selector
    seconds: float = 1.0

    def execute(self, workflow):
        while not workflow.js.element_exists(self.selector):
            sleep(self.seconds)


@dataclass(frozen=True)
class WaitForUser(PageAction):
    """
    Waits until the Enter key is pressed in the terminal.
    """

    def execute(self, _):
        _ = input("Click [Enter] to continue...")


@dataclass(frozen=True)
class Refresh(PageAction):
    """
    Triggers a refresh of the current page.
    Note that the following snapshot may not have wtl_uid that
    map equally to the previous state.
    """

    def execute(self, workflow):
        workflow.scraper.refresh()


@dataclass(frozen=True)
class Abort(PageAction):
    """
    Stops any future progress on this tab and will set an aborted flag
    on the given tab. The tab will not be snapshoted in the future.
    If all tabs have received an Abort call, the workflow will stop.
    Note! If you're using multiple tabs, this action has highest priority.
    """

    def execute(self, workflow):
        if workflow.config.actions.abort.close and not workflow.config.debug.preserve_window:
            workflow.current_window.close_tab()



"""
Helper code to check versions of installed dependencies.
"""

import logging
from subprocess import CalledProcessError, run

logger = logging.getLogger("wtl")


class VERSION_CMD:
    """Terminal strings for getting version info"""

    CHROME = "google-chrome --version"
    CHROMIUM = "chromium --version"
    CHROMEDRIVER = "chromedriver --version"
    FIREFOX = "firefox --version"
    GECKODRIVER = "geckodriver --version"


def run_cmd(cmd: str, title: str = None) -> bool:
    """Runs given command. Outputs data if descriptive title is given"""
    try:
        result = run(cmd.split(" "), check=True, capture_output=True)
        if title:
            logger.info(f"{title}: {str(result.stdout.decode('ascii').strip())}")
        return True
    except CalledProcessError:
        if title:
            logger.info(f"{title}: [FAILED]")

    return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Module containing helper functions for graphics-related operations on webdrivers and snapshots.
"""

import importlib.resources
import logging

from PIL import Image, ImageDraw, ImageFont

from .color import Color
from .geometry import Point, Rectangle

logger = logging.getLogger("wtl")


def crop_image(image: Image.Image, rect: Rectangle) -> Image.Image:
    """
    Crops the part of the image specified by its ``rect``.

    Rectangle specified by ``rect`` must lie inside of the image bounds.
    """

    if rect.area == 0:
        raise ValueError(f"Rectangle {rect} is degenerate")

    # Rectangle we crop out must be somewhere within the image
    image_rect = Rectangle(Point(0, 0), Point(image.width, image.height))
    if rect not in image_rect:
        raise ValueError(f"Bounds {rect} outside of image area {image_rect}")

    return image.crop(rect.bounds)


def draw_rect(image: Image.Image, rect: Rectangle, color: Color, width: int):
    """
    Draws a bounding box around the specified rectangle on the image.
    """
    draw = ImageDraw.Draw(image, mode="RGBA")
    draw.rectangle(rect.bounds, outline=color.to_tuple(with_alpha=True), width=width)


def draw_text(image: Image.Image, top_left: Point, color: Color, size: int, text: str):
    """
    Draws text on a PIL image.
    """
    padding = 2
    image_width, image_height = image.size

    with importlib.resources.path("webtraversallibrary.font", "OpenSans-Regular.ttf") as filepath:
        font = ImageFont.truetype(str(filepath), size)

    # Make sure text does not exceed image boundaries
    text_width, text_height = font.getsize(text)
    draw_text_x = min(top_left.x, image_width - text_width - padding)
    draw_text_y = min(top_left.y, image_height - text_height - padding)
    draw = ImageDraw.Draw(image, mode="RGB")
    draw.text(xy=(draw_text_x, draw_text_y), text=text, fill=color.to_tuple(), font=font)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Abstraction layer for interactions with a browser instance."""
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Set

from selenium.common.exceptions import JavascriptException, UnexpectedAlertPresentException, WebDriverException

from .config import Config
from .error import WindowClosedError
from .javascript import JavascriptWrapper
from .logging import logging
from .scraper import Scraper
from .webdrivers import setup_driver

logger = logging.getLogger("wtl")


class Window:
    """
    Owns a webdriver, scraper and javascript wrapper.
    Handles browser instance logic.
    """

    def __init__(
        self, config: Config, preload_callbacks: Iterable[Path] = None, postload_callbacks: List[Callable] = None
    ):
        self._driver = setup_driver(config, preload_callbacks=preload_callbacks)
        self.scraper = Scraper(driver=self.driver, config=config, postload_callbacks=postload_callbacks)
        self.js = JavascriptWrapper(self.driver, config)
        self.name_to_handle: Dict[str, int] = {}
        self.closed: Set[str] = set()
        self.current: str = None
        self.tab_navigation: Dict[str, str] = {}

    def ensure_running(self):
        """
        Tries fetching attached window handles.
        Raises WindowClosedError if this fails.
        """
        try:
            assert self._driver
            _ = self._driver.window_handles
        except (AssertionError, WebDriverException) as e:
            raise WindowClosedError("Could not reach the browser window!") from e

    @property
    def driver(self):
        """Checks if the driver is attached, and if so returns a reference to it."""
        self.ensure_running()
        return self._driver

    def is_closed(self, tab):
        """Returns True if the tab has been closed"""
        return tab in self.closed

    @property
    def tabs(self):
        """Returns a list of all tab names"""
        return list(self.name_to_handle.keys())

    @property
    def open_tabs(self):
        """Returns a list of all open tab names"""
        return [tab for tab in self.tabs if not self.is_closed(tab)]

    @property
    def navigation(self):
        """
        Returns URL to navigate to for the current active tab, if it exists.
        Removes the entry afterwards, to two subsequent calls need not be equal.
        """
        if self.current not in self.tab_navigation:
            return None

        result = self.tab_navigation[self.current]
        del self.tab_navigation[self.current]
        return result

    @property
    def cookies(self) -> List[Any]:
        return self.driver.get_cookies()

    @cookies.setter
    def cookies(self, value: Iterable[Any]):
        """Replaces(!) all existing cookies with the given list."""
        self.driver.delete_all_cookies()
        for cookie in value:
            self.driver.add_cookie(cookie)

    def create_tab(self, name: str, url: str = "about:blank"):
        """
        Open a new tab with a given `name` (must be unique for this Window).
        If `url` argument is given, saves this value that can be retrieved using the
        :func:`navigation` method later. The tab will not automatically navigate
        to the given adress. Use a :class:`Scraper` or interact directly with the driver.
        """
        assert self.driver, "Cannot interact with window after calling quit!"
        assert name not in self.closed, "Cannot reopen a closed tab!"

        # Every window comes with one tab, reuse if possible
        if len(self.name_to_handle) > 0:
            self.js.execute_script("window.open('about:blank');")

        # Get mapping of the new tab and store in the dict
        for _tab in self.driver.window_handles:
            if _tab not in self.name_to_handle.values():
                self.name_to_handle[name] = _tab

        # Remember navigation
        if url != "about:blank":
            self.tab_navigation[name] = url

    def set_tab(self, name: str):
        """
        Sets the current active tab to the one listed under the given `name`.
        Throws AssertionError if no such tab exists.

        .. warning::
            If the tab has been closed, interactions may lead to weird behaviour.
        """
        assert name in self.name_to_handle, f"No such tab exists: {name}"
        self.current = name

        if name in self.closed:
            logger.warning("This tab has been closed. Do not interact with it.")
        else:
            self.driver.switch_to.window(self.name_to_handle[name])

    def close_tab(self):
        """Closes the current active tab."""
        assert self.current not in self.closed, "Closing already closed tab!"
        self.closed.add(self.current)
        try:
            self.driver.close()
        except (UnexpectedAlertPresentException, JavascriptException, WebDriverException, WindowClosedError):
            logger.warning("Attempting to close failed - window probably already closed.")

    def quit(self):
        """
        Terminates the browser and associated driver of this instance.
        Do not use this window afterwards.
        """
        if not self._driver:
            return

        try:
            for tab in self.open_tabs:
                self.set_tab(tab)
                self.close_tab()
            self._driver.quit()
        except WindowClosedError:
            logger.warning("Attempting to close already closed window.")
        finally:
            self._driver = None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Collection of helper classes used in Workflow.
"""

import logging
from collections.abc import Collection
from typing import Any, Dict, Iterable, Union

from selenium import webdriver

from .classifiers import Classifier, ElementClassifier, ViewClassifier
from .error import ElementNotFoundError
from .javascript import JavascriptWrapper
from .selector import Selector
from .snapshot import PageElement, PageSnapshot

logger = logging.getLogger("wtl")


class ClassifierCollection(Collection):
    """Helper class for predefined classifiers"""

    def __init__(self, classifiers: Iterable[Classifier]):
        self._classifiers: Dict[str, Classifier] = {}
        if classifiers:
            for classifier in classifiers:
                self.add(classifier)

    def add(self, classifier: Classifier):
        self._classifiers[classifier.name] = classifier

    def start(self, classifier: Union[str, Classifier]):
        name = classifier if isinstance(classifier, str) else classifier.name
        self._classifiers[name].enabled = True

    def stop(self, classifier: Union[str, Classifier]):
        name = classifier if isinstance(classifier, str) else classifier.name
        self._classifiers[name].enabled = False

    @property
    def active_element_classifiers(self):
        return ClassifierCollection([c for c in self if c.enabled and c.callback and isinstance(c, ElementClassifier)])

    @property
    def active_view_classifiers(self):
        return ClassifierCollection([c for c in self if c.enabled and c.callback and isinstance(c, ViewClassifier)])

    def __iter__(self):
        yield from self._classifiers.values()

    def __contains__(self, classifier: Any):
        if isinstance(classifier, str):
            return classifier in self._classifiers
        if isinstance(classifier, Classifier):
            return classifier in self._classifiers.values()
        raise TypeError

    def __len__(self):
        return len(self._classifiers)


class MonkeyPatches:
    """Helper class for monkeypatches"""

    def __init__(self, patches: Dict[Selector, str] = None):
        self._data: Dict[Selector, str] = patches or {}
        self._default: str = None

    def add(self, selector: Selector, patch: str):
        self._data[selector] = patch

    def set_default(self, patch: str):
        """Equivalent to ``check(Selector("*"), element)`` but much faster."""
        self._default = patch

    def check(self, snapshot: PageSnapshot, element: PageElement) -> str:
        """If a rule applies for given element for given snapshot, return the most specific value"""
        selector_elements = [(s, snapshot.elements.by_selector(s)) for s in self._data]
        selector_elements.sort(key=lambda item: len(item[1]), reverse=True)
        for selector, elements in selector_elements:
            if element in elements:
                return self._data[selector]
        return self._default

    def __contains__(self, selector: Selector):
        return selector in self._data

    def __len__(self):
        return len(self._data)


class FrameSwitcher:
    """
    Helper class for entering and exiting iframes.
    Raises ElementNotFoundError if an iframe could not be found.
    """

    def __init__(self, identifier: str, js: JavascriptWrapper, driver: webdriver):
        self.name = None
        self.driver = driver

        if identifier:
            self.name = js.find_iframe_name(identifier)
            if not self.name:
                raise ElementNotFoundError(f"Found no iframe with identifier '{identifier}'")

    def __enter__(self):
        """
        Goes into an <iframe> object with given name.
        """
        if self.name:
            logger.debug(f"Entering iframe: '{self.name}'")
            self.driver.switch_to.frame(self.name)

    def __exit__(self, *_):
        """
        Steps out into the parent frame.
        """
        if self.name:
            logger.debug(f"Exiting iframe: '{self.name}'")
            self.driver.switch_to.default_content()
